#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class InspectError(Exception):

    "An inspection error."

    pass

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __getitem__(self, name):
        return self.namespace[name]

    def keys(self):
        return self.namespace.keys()

class Class(NamespaceDict):

    "An inspected class."

    def __init__(self, name):
        self.name = name
        self.bases = []
        self.namespace = {}
        self.instattr = set()

    def add_base(self, base):
        self.bases.append(base)

class Function(NamespaceDict):

    "An inspected function."

    def __init__(self, name, argnames, has_star, has_dstar):
        self.name = name
        self.argnames = argnames
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.namespace = {}

class UnresolvedName(NamespaceDict):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name):
        self.name = name
        self.namespace = {}

class Module(ASTVisitor, NamespaceDict):

    "An inspected module."

    def __init__(self, name, importer=None):
        ASTVisitor.__init__(self)
        self.visitor = self

        self.name = name
        self.importer = importer
        self.loaded = 0

        # Module namespace.

        self.namespace = {}

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_global = 1
        self.in_init = 0
        self.classes = []
        self.module = None

    def parse(self, filename):
        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):
        self.module = module
        processed = self.dispatch(module)
        if self.namespace.has_key("__all__"):
            all = self.namespace["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self.namespace[n.value] = self.importer.add_module(self.name + "." + n.value)
        return processed

    def vacuum(self):
        for name, value in self.namespace.items():
            if isinstance(value, Module) and not value.loaded:
                del self.namespace[name]

    # Namespace methods.

    def store(self, name, obj):
        if self.in_global:
            if not self.classes:
                self.namespace[name] = obj
            else:
                self.classes[-1].namespace[name] = obj

    def store_attr(self, name):
        if self.in_init:
            self.classes[-1].instattr.add(name)

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError, node.__class__

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def NOP(self, node):
        return node

    visitAdd = NOP

    visitAnd = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None and isinstance(expr, Self):
            self.store_attr(node.attrname)
        return None

    def visitAssList(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssName(self, node):
        self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    visitAugAssign = NOP

    visitBitand = NOP

    visitBitor = NOP

    visitCallFunc = NOP

    def visitClass(self, node):
        if not self.in_global:
            raise InspectError, "Class is %s not global: cannot handle this reasonably." % node.name
        else:
            cls = Class(node.name)
            for base in node.bases:
                base_ref = self.dispatch(base)
                if base_ref is None:
                    raise InspectError, "Base class %s for class %s is not found: it may be hidden in some way." % (base, node.name)
                cls.add_base(base_ref)

            # Make an entry for the class.

            self.store(node.name, cls)

            self.classes.append(cls)
            self.dispatch(node.code)
            self.classes.pop()

        return node

    visitConst = NOP

    visitDict = NOP

    visitDiscard = NOP

    visitFor = NOP

    def visitFrom(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'from' statement."

        module = self.importer.load(node.modname, 1)

        if module is None:
            print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None:
                    self.namespace[alias or name] = attr = module.namespace[name]
                    if isinstance(attr, Module) and not attr.loaded:
                        self.importer.load(attr.name)
                else:
                    self.namespace[alias or name] = UnresolvedName(name)
            else:
                if module is not None:
                    for n in module.namespace.keys():
                        self.namespace[n] = attr = module.namespace[n]
                        if isinstance(attr, Module) and not attr.loaded:
                            self.importer.load(attr.name)

        return None

    def visitFunction(self, node):
        function = Function(
            node.name,
            node.argnames,
            has_star=(node.flags & 4 != 0),
            has_dstar=(node.flags & 8 != 0)
            )

        in_global = self.in_global
        self.in_global = 0
        if node.name == "__init__" and self.classes:
            self.in_init = 1
            self.dispatch(node.code)
            self.in_init = 0
        self.in_global = in_global

        self.store(node.name, function)
        return None

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None and isinstance(expr, Module):
            return expr.namespace.get(node.attrname)
        else:
            return builtins_namespace.get(node.attrname)

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    def visitImport(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'import' statement."

        for name, alias in node.names:
            if alias is not None:
                self.namespace[alias] = self.importer.load(name, 1)
            else:
                self.namespace[name.split(".")[0]] = self.importer.load(name)

        return None

    visitLeftShift = NOP

    visitList = NOP

    def visitModule(self, node):
        return self.dispatch(node.node)

    visitMul = NOP

    visitMod = NOP

    def visitName(self, node):
        name = node.name
        if name == "self":
            return Self()
        elif self.namespace.has_key(name):
            return self.namespace[name]
        elif builtins_namespace.has_key(name):
            return builtins_namespace[name]
        else:
            return None

    visitNot = NOP

    visitOr = NOP

    visitPass = NOP

    visitRaise = NOP

    visitRightShift = NOP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSubscript = NOP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:
            self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitTuple = NOP

class Self:

    "A reference to an object within a method."

    pass

builtins_namespace = {}
for key in ['ArithmeticError', 'AssertionError', 'AttributeError',
    'BaseException', 'DeprecationWarning', 'EOFError', 'Ellipsis',
    'EnvironmentError', 'Exception', 'False', 'FloatingPointError',
    'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError',
    'ImportWarning', 'IndentationError', 'IndexError', 'KeyError',
    'KeyboardInterrupt', 'LookupError', 'MemoryError', 'NameError',
    'None', 'NotImplemented', 'NotImplementedError', 'OSError',
    'OverflowError', 'PendingDeprecationWarning', 'ReferenceError',
    'RuntimeError', 'RuntimeWarning', 'StandardError', 'StopIteration',
    'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit',
    'TabError', 'True', 'TypeError', 'UnboundLocalError',
    'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
    'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning',
    'ValueError', 'Warning', 'ZeroDivisionError', 'object']:
    builtins_namespace[key] = Class(key)

# vim: tabstop=4 expandtab shiftwidth=4
